#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import dbus
import dbus.service
import sys
import os
import json
from installhandler import InstallHandler, Status
import logging

import subprocess
from subprocess import Popen, PIPE

import commands
import time
# from apscheduler.schedulers.background import BackgroundScheduler
from datetime import datetime
from threading import Timer

# set dbus msg
global dbus_name
global dbus_path
global dbus_iface
dbus_name = "org.cdos.update"
dbus_path = "/org/cdos/update"
dbus_iface =  "org.cdos.update"

patchmgr_app_path = "/usr/share/cdos-update"
patchmgr_patches_dir = patchmgr_app_path + "/updates"
kernel_live_patches_dir = patchmgr_app_path + "/kernel_live_patches"
patches_tmp_dir = "/tmp/cdosupdate"

PATCHFILE_EXT_NAME = "cdospatch"

status_file_path = "/usr/share/cdos-update/daemon/status"
time_file_path = "/usr/share/cdos-update/daemon/time"

logging.basicConfig(format= '%(levelname)s %(asctime)-15s %(message)s file:%(filename)s line:%(lineno)d', filename='/var/log/cdosupdate-daemon.log',level=logging.DEBUG)

# def checkUpdateStatus():
    # currentTime = datetime.now().strftime("%Y%m%d%H%M%S")
    # status, output = commands.getstatusoutput("check-update-status")
    # if (0 != status):
    #     logging.debug(datetime.now().strftime("%Y/%m/%d %H:%M:%S: ") + "check update status: Error")
    # else:
    #     logging.debug(datetime.now().strftime("%Y/%m/%d %H:%M:%S: ") + "check update status: Success")
    # t = Timer(7200, checkUpdateStatus)
    # t.start()

    # 如果失败，10分钟后重新执行
    # while(0 != status):
    #     logging.debug("checkUpdateStatus: reoperating")
    #     time.sleep(600)
    #     status, output = commands.getstatusoutput("check-update-status")

def SendData():
    status, output = commands.getstatusoutput("check-update-status")
    if (0 != status):
        print("Error: " + output)
    else:
        print("Success")

    statusFile = open(status_file_path, 'w')
    statusFile.write(str(status))
    statusFile.close()

def CheckUpdateStatusPerHour():
    print("CheckUpdateStatusPerHour")

    # 1小时之后继续检查
    t = Timer(3600, CheckUpdateStatusPerHour)

    currentDate = datetime.now().strftime("%Y%m%d")

    # 首次执行
    if (not os.path.isfile(time_file_path)):
        # print("文件不存在，创建并写入日期，发送数据")
        logging.debug(datetime.now().strftime("%Y/%m/%d %H:%M:%S") + ": 文件不存在，创建并写入日期，发送数据")
        timeFile = open(time_file_path, "w")
        timeFile.write(currentDate)
        timeFile.close()

        SendData()
        t.start()
        return

    # 判断上次是否执行成功
    statusFile = open(status_file_path, "r")
    statusOfFile = statusFile.read()
    statusOfFile = statusOfFile.strip('\n')
    statusFile.close()
    if (statusOfFile and "0" != statusOfFile):
        # print("上次执行失败，重新执行")
        logging.debug(datetime.now().strftime("%Y/%m/%d %H:%M:%S") + ": 上次执行失败，重新执行")
        SendData()
        t.start()
        return

    # 判断是否是同一天
    timeFile = open(time_file_path, "r")
    dateOfFile = timeFile.read();
    dateOfFile = dateOfFile.strip('\n')
    timeFile.close()
    # 日期不一致，写入新日期并发送数据
    if (currentDate != dateOfFile):
        # print("日期不一致，发送数据" + datetime.now().strftime("%Y%m%d%H%M"))
        logging.debug(datetime.now().strftime("%Y/%m/%d %H:%M:%S") + ": 日期不一致，写入新日期并发送数据")
        SendData()
        timeFile = open(time_file_path, "w")
        timeFile.write(currentDate)
        timeFile.close()

        t.start()

        return

    # print("日期一致，不发送" + datetime.now().strftime("%Y%m%d%H%M"))
    logging.debug(datetime.now().strftime("%Y/%m/%d %H:%M:%S") + ": 日期一致，不发送")
    t.start()

class PatchInstallerObject(dbus.service.Object):

    def __init__(self):
        self.session_bus = dbus.SystemBus()
        name = dbus.service.BusName(dbus_name, bus=self.session_bus)
        dbus.service.Object.__init__(self, name, dbus_path)
        self._status = Status.INITED
        self._installedPatches = []

        self._installHandler = InstallHandler()
        self.checkInstalled()
        self.installKernelPatches()

        CheckUpdateStatusPerHour()

    @dbus.service.method(dbus_iface, in_signature='', out_signature='')
    def installKernelPatches(self):
        logging.debug("installKernelPatches")
        # 初始化热补丁目录
        if os.path.exists(kernel_live_patches_dir) and os.path.isfile(kernel_live_patches_dir):
            os.remove(kernel_live_patches_dir)
        if not os.path.exists(kernel_live_patches_dir):
            os.makedirs(kernel_live_patches_dir, 0644)
        deb_info = self.get_deb_info()
        if deb_info == {}:
            return
        # 获取已安装列表
        cmd = "kpatch list|grep kpatch"
        subp = Popen(cmd, shell=True, stdout=PIPE)
        rc = subp.wait()
        installed_kernel_ids_str = subp.stdout.read().strip().split('\n')
        # 已安装列表
        installed_kernel_ids = []
        for installed_kernel_id_str in installed_kernel_ids_str:
            if not ( installed_kernel_id_str == ''):
                installed_kernel_ids.append( installed_kernel_id_str[7:])
        # 获取本地待安装内核热补丁
        for kernel_live_patch in os.listdir(kernel_live_patches_dir):
            # 判断内核热补丁合法性
            kernel_live_patch_path = os.path.join(kernel_live_patches_dir, kernel_live_patch)
            kernel_live_patch_path_json = kernel_live_patch_path+".json"
            if os.path.isfile(kernel_live_patch_path) and os.path.isfile(kernel_live_patch_path_json) and kernel_live_patch.startswith("kpatch-") and kernel_live_patch.endswith(".ko"):
                try:
                    # 获取内核热补丁id
                    kernel_id = kernel_live_patch[7:-3]
                    # 如果该热补丁未被安装
                    if (kernel_id not in installed_kernel_ids) and (self.check_kernel(kernel_live_patch_path_json, deb_info) == True):
                        # 安装待安装的内核热补丁
                        os.system("sudo kpatch load " + kernel_live_patch_path)
                except Exception as e:
                    logging.error('Exception %s:  %s' % (e, traceback.format_exc()))
                    continue
        return

    def get_deb_info(self):
        # 获取本地已经正确安装的deb包及版本的json描述(即ii状态的deb包)
        deblist = subprocess.check_output(["dpkg","-l"]).split('\n')
        if len(deblist) <= 5:
            return {}
        deb_desc_json = {}
        for i in range(5,len(deblist) - 1):
            deb_desc = deblist[i].split()
            if (len(deb_desc) > 0) and (deb_desc[0] == "ii"):
                deb_desc_json[deb_desc[1]] = deb_desc[2]
        return deb_desc_json

    def check_kernel(self, kernel_json, deb_desc_json):
        try:
            kernel_json_file = file(kernel_json)
            kernel_info = json.load(kernel_json_file)
            os_kernel_release = os.uname()[2]
            if not os_kernel_release == kernel_info["kernel_release"]:
                return False
            # 开始校验包版本
            package_info = kernel_info["packages"]
            for deb_name in package_info:
                # 如果本地deb包中没有该包
                if deb_name not in deb_desc_json:
                    return False
                if not type(package_info[deb_name]) == list :
                    return False
                if len(package_info[deb_name]) != 0 and deb_desc_json[deb_name] not in package_info[deb_name]:
                    return False
            return True
        except Exception as e:
            return False

    @dbus.service.method(dbus_iface, in_signature='', out_signature='s')
    def getVersion(self):
        logging.debug("getVersion 1.1.3")
        return "1.1.3"

    @dbus.service.method(dbus_iface, in_signature='', out_signature='i')
    def checkInstalled(self):
        logging.debug("checkInstalled")
        try:
            alist = []
            if not os.path.exists(patchmgr_patches_dir):
                os.makedirs(patchmgr_patches_dir, 0644)
            for afile in os.listdir(patchmgr_patches_dir):
                if os.path.isfile(os.path.join(patchmgr_patches_dir, afile)):
                    if afile.endswith("." + PATCHFILE_EXT_NAME):
                        alist.append(afile[:-(1+len(PATCHFILE_EXT_NAME))])
            self._installedPatches = alist
        except Exception as e:
            logging.error('Exception %s:  %s' % (e, traceback.format_exc()))
            return -1
        return len(self._installedPatches)

    @dbus.service.method(dbus_iface, in_signature='', out_signature='as')
    def getInstalledList(self):
        logging.debug("getInstalledList")
        try:
            alist = []
            if not os.path.exists(patchmgr_patches_dir):
                return alist
            for afile in os.listdir(patchmgr_patches_dir):
                if os.path.isfile(os.path.join(patchmgr_patches_dir, afile)):
                    if afile.endswith("." + PATCHFILE_EXT_NAME):
                        alist.append(afile[:-(1+len(PATCHFILE_EXT_NAME))])
            return alist
        except Exception as e:
            logging.error('Exception %s:  %s' % (e, traceback.format_exc()))
        return []

    @dbus.service.method(dbus_iface, in_signature='', out_signature='as')
    def getDownloadedList(self):
        logging.debug("getDownloadedList")
        downloadedList = []
        try:
            if not os.path.exists(patches_tmp_dir):
                return downloadedList
            for afile in os.listdir(patches_tmp_dir):
                if os.path.isfile(os.path.join(patches_tmp_dir, afile)):
                    if afile.endswith("." + PATCHFILE_EXT_NAME):
                        downloadedList.append(afile[:-(1+len(PATCHFILE_EXT_NAME))])
            return downloadedList
        except Exception as e:
            logging.error('Exception %s:  %s' % (e, traceback.format_exc()))
        return downloadedList

    def progressCallback(self, statusCode, status_text, patchID, correct_install_patches_list, patchList, progress, paramList={}):
        logging.info("progressCallback")
        if statusCode == Status.INSTALLED:
            self._installedPatches.append(patchID)
        if statusCode == Status.ALLINSTALLED:
            self.checkInstalled()
        if statusCode == Status.ERROR_INSTALL:
            self.checkInstalled()
        self.statusChanged(statusCode, status_text, patchID, correct_install_patches_list, patchList, progress, paramList)

    @dbus.service.method(dbus_iface, in_signature='sas', out_signature='s')
    def download(self, getpatch_site_path, patchIDs):
        logging.info("download: %s %s" % (getpatch_site_path, str(patchIDs)) )
        if self._status == Status.INITED or self._status == Status.ALLDOWNLOADED or self._status == Status.ALLINSTALLED or self._status == Status.ERROR_INSTALL:
            patchIDList = []
            for i in range(len(patchIDs)):
                patchIDList.append(str(patchIDs[i]))
            try:
                if self._installHandler.start_Update(getpatch_site_path, patchIDList, self.progressCallback, True):
                    return "Started"
            except Exception, e:
                logging.error('Exception %s:  %s' % (e, traceback.format_exc()))
            return "Failed"
        else:
            return "Failed"

    @dbus.service.method(dbus_iface, in_signature='sas', out_signature='s')
    def install(self, getpatch_site_path, patchIDs):
        logging.info("install: %s %s" % (getpatch_site_path, str(patchIDs)) )
        if self._status == Status.INITED or self._status == Status.ALLDOWNLOADED or self._status == Status.ALLINSTALLED or self._status == Status.ERROR_INSTALL:
            patchIDList = []
            for i in range(len(patchIDs)):
                patchIDList.append(str(patchIDs[i]))
            try:
                if self._installHandler.start_Update(getpatch_site_path, patchIDList, self.progressCallback):
                    return "Started"
            except Exception, e:
                logging.error('Exception %s:  %s' % (e, traceback.format_exc()))
            return "Failed"
        else:
            return "Failed"

    @dbus.service.method(dbus_iface, in_signature='', out_signature='i')
    def getStatus(self):
        logging.debug("getStatus:" + self._status)
        return self._status

    @dbus.service.signal(dbus_interface=dbus_iface, signature='issasasia{sv}',)
    def statusChanged(self, statusCode, status_text, patchID, correct_install_patches_list, patchList, progress, paramList={}):
        logging.debug("statusChanged:" + str(statusCode) + " " + status_text + " " + patchID + " " + str(correct_install_patches_list) + " " + str(patchList) + " " + str(progress) + " " + str(paramList))

if __name__ == '__main__':
    # using glib
    import dbus.mainloop.glib
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
    from gi.repository import GObject
    GObject.threads_init()
    loop = GObject.MainLoop()
    patchInstaller = PatchInstallerObject()
    loop.run()
